import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from click.testing import CliRunner
from nose.tools import ok_
from numpy.testing import assert_equal

from neuror.cli import cli
from neuror.full import full

from .test_full import assert_output_exists

DATA = Path(__file__).parent / 'data'


def test_full_custom_plots_dir():
    try:
        from neuror.view import plot_repaired_neuron
    except ImportError:
        print('Skipping this test as [plotly] extra is not installed here')
        return

    with TemporaryDirectory('test-cli-full') as tmp_folder:
        test_folder = Path(tmp_folder, 'test-full-repair')
        shutil.copytree(DATA / 'test-full-repair', test_folder)

        custom_path = test_folder / 'plots_custom'
        full(test_folder, plots_dir=custom_path)
        assert_output_exists(test_folder)
        assert_equal(len(list(custom_path.iterdir())), 3)
        ok_((custom_path / 'report.pdf').exists())


def test_report():
    try:
        from neuror.view import plot_repaired_neuron
    except ImportError:
        print('Skipping this test as [plotly] extra is not installed here')
        return

    with TemporaryDirectory('test-report') as folder:
        runner = CliRunner()
        result = runner.invoke(cli, ['report', str(DATA), folder])
        assert_equal(result.exit_code, 0)



import shutil
import sys
from pathlib import Path
from tempfile import TemporaryDirectory

from mock import patch
from nose.tools import assert_raises, ok_
from numpy.testing import assert_equal

from neuror.full import full

DATA_PATH = Path(__file__).parent / 'data'


def assert_output_exists(root_dir,
                         raw_dir=None,
                         raw_planes_dir=None,
                         unravelled_dir=None,
                         unravelled_planes_dir=None,
                         repaired_dir=None):
    raw_dir = raw_dir or Path(root_dir, 'raw')
    raw_planes_dir = raw_planes_dir or Path(raw_dir, 'planes')
    unravelled_dir = unravelled_dir or Path(root_dir, 'unravelled')
    unravelled_planes_dir = unravelled_planes_dir or Path(unravelled_dir, 'planes')
    repaired_dir = repaired_dir or Path(root_dir, 'repaired')

    for folder in [raw_dir, raw_planes_dir, unravelled_dir, unravelled_planes_dir, repaired_dir]:
        ok_(folder.exists(), '{} does not exists'.format(folder))
        ok_(list(folder.iterdir()), '{} is empty !'.format(folder))


def test_full():
    with TemporaryDirectory('test-full') as tmp_folder:
        test_folder = Path(tmp_folder, 'test-full-repair')
        shutil.copytree(DATA_PATH / 'test-full-repair', test_folder)
        full(test_folder)
        assert_output_exists(test_folder)


def test_full_custom_raw_dir():
    with TemporaryDirectory('test-full-custom-raw-dir') as tmp_folder:
        test_folder = Path(tmp_folder, 'test-full-repair')
        shutil.copytree(DATA_PATH / 'test-full-repair', test_folder)

        raw_dir_custom_path = test_folder / 'raw_custom'
        # Should raise because raw_custom dir does not exist
        assert_raises(Exception, full, test_folder, raw_dir=raw_dir_custom_path)

        shutil.move(test_folder / 'raw', raw_dir_custom_path)
        full(test_folder, raw_dir=raw_dir_custom_path)
        assert_output_exists(test_folder, raw_dir=raw_dir_custom_path)


def test_full_custom_unravel_dir():
    with TemporaryDirectory('test-full-custum-unravel-dir') as tmp_folder:
        test_folder = Path(tmp_folder, 'test-full-repair')
        shutil.copytree(DATA_PATH / 'test-full-repair', test_folder)

        custom_path = test_folder / 'unravel_custom'
        full(test_folder, unravelled_dir=custom_path)
        assert_output_exists(test_folder, unravelled_dir=custom_path)


def test_full_custom_unravelled_planes_dir():
    with TemporaryDirectory('test-full-custom-unravelled-planes-dir') as tmp_folder:
        test_folder = Path(tmp_folder, 'test-full-repair')
        shutil.copytree(DATA_PATH / 'test-full-repair', test_folder)

        custom_path = test_folder / 'unravelled_planes_custom'
        full(test_folder, unravelled_planes_dir=custom_path)
        assert_output_exists(test_folder, unravelled_planes_dir=custom_path)


def test_full_custom_repaired_dir():
    with TemporaryDirectory('test-full-custom-repaired-dir') as tmp_folder:
        test_folder = Path(tmp_folder, 'test-full-repair')
        shutil.copytree(DATA_PATH / 'test-full-repair', test_folder)

        custom_path = test_folder / 'repaired_planes_custom'
        full(test_folder, repaired_dir=custom_path)
        assert_output_exists(test_folder, repaired_dir=custom_path)



'''Dendritic repair module

It is based on the BlueRepairSDK's implementation
'''
import logging
from collections import Counter, OrderedDict, defaultdict
from enum import Enum
from itertools import chain

import neurom as nm
import numpy as np
from morph_tool import apical_point_section_segment
from morphio import PointLevel, SectionType
from neurom import NeuriteType, iter_neurites, iter_sections, load_neuron
from neurom.core.dataformat import COLS
from neurom.features.sectionfunc import branch_order, section_path_length
from scipy.spatial.distance import cdist

from neuror import axon
from neuror.cut_plane import CutPlane
from neuror.utils import direction, rotation_matrix, section_length

SEG_LENGTH = 5.0
SHOLL_LAYER_SIZE = 10
BIFURCATION_BOOSTER = 0
NOISE_CONTINUATION = 0.7
SOMA_REPULSION = 0.7
BIFURCATION_ANGLE = 0

# Epsilon needs not to be to small otherwise leaves stored in json files
# are not found in the NeuroM neuron
EPSILON = 1e-6

L = logging.getLogger('neuror')


class Action(Enum):
    '''To bifurcate or not to bifurcate ?'''
    BIFURCATION = 1
    CONTINUATION = 2
    TERMINATION = 3


class RepairType(Enum):
    '''The types used for the repair

    based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.h#n22  # noqa, pylint: disable=line-too-long
    '''
    trunk = 0
    tuft = 1
    oblique = 2
    basal = 3
    axon = 4


def is_cut_section(section, cut_points):
    '''Return true if the section is close from the cut plane'''
    if cut_points.size == 0 or section.points.size == 0:
        return False
    return np.min(cdist(section.points[:, COLS.XYZ], cut_points)) < EPSILON


def is_branch_intact(branch, cut_points):
    '''Does the branch have leaves belonging to the cut plane ?'''
    return all(not is_cut_section(section, cut_points)
               for section in branch.ipreorder())


def _get_sholl_layer(section, origin):
    '''Returns this section sholl layer'''
    return int(np.linalg.norm(section.points[-1, COLS.XYZ] - origin) / SHOLL_LAYER_SIZE)


def _get_sholl_proba(sholl_data, section_type, sholl_layer, pseudo_order):
    '''Return the probabilities of bifurcation, termination and bifurcation
    in a dictionnary for the given sholl layer and branch order.

    If no data are available for this branch order, the action_counts
    are averaged on all branch orders for this sholl layer


    Args:
        sholl_data: nested dict that stores the number of section per SectionType, sholl order
            sholl layer and Action type
            sholl_data[neurite_type][layer][order][action_type] = counts
        section_type (SectionType): section type
        sholl_layer (int): sholl layer
        pseudo_order (int): pseudo order

    Returns:
        Dict[Action, float]: probability of each action

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n398  # noqa, pylint: disable=line-too-long
    Note2: OrderedDict ensures the reproducibility of np.random.choice outcome
    '''

    section_type_data = sholl_data[section_type]
    try:
        data_layer = section_type_data[sholl_layer]
    except KeyError:
        return OrderedDict([(Action.BIFURCATION, 0),
                            (Action.CONTINUATION, 0),
                            (Action.TERMINATION, 1)])

    try:
        action_counts = data_layer[pseudo_order]
    except KeyError:
        # No data for this order. Average on all orders for this layer
        # As done in https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n426  # noqa, pylint: disable=line-too-long

        action_counts = Counter()
        for data in data_layer.values():
            for action, count in data.items():
                action_counts[action] += count

    total_counts = sum(action_counts.values())
    if total_counts == 0:
        return OrderedDict([(Action.BIFURCATION, 0),
                            (Action.CONTINUATION, 0),
                            (Action.TERMINATION, 1)])

    boost_bifurcation = int(total_counts * BIFURCATION_BOOSTER)
    action_counts[Action.BIFURCATION] += boost_bifurcation
    total_counts += boost_bifurcation
    res = OrderedDict([tuple((action, action_counts[action] / float(total_counts))) for action in
                       sorted(action_counts.keys(), key=lambda t: t.value)])
    return res


def _grow_until_sholl_sphere(section, origin, sholl_layer):
    '''Grow until reaching next sholl layer

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n363  # noqa, pylint: disable=line-too-long
    '''
    backwards_sections = 0
    while _get_sholl_layer(section, origin) == sholl_layer and backwards_sections < 4:
        _continuation(section, origin)

        # make sure we don't grow back the origin
        is_last_segment_toward_origin = np.dot(
            section.points[-1, COLS.XYZ] - origin,
            _last_segment_vector(section)) < 0

        if is_last_segment_toward_origin:
            backwards_sections += 1
    return backwards_sections


def _last_segment_vector(section, normalized=False):
    '''Returns the vector formed by the last 2 points of the section'''
    vec = section.points[-1, COLS.XYZ] - section.points[-2, COLS.XYZ]
    if normalized:
        return vec / np.linalg.norm(vec)
    return vec


def _get_similar_child_diameters(sections, original_section):
    '''Find an existing section with a similar diameter and returns the diameters
    of its children

    If no similar section is found, returns twice the diameter of the last point

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n76  # noqa, pylint: disable=line-too-long
    '''
    threshold = 0.2
    diameter = original_section.points[-1, COLS.R] * 2

    for section in sections:
        if section.children and abs(section.points[-1, COLS.R] * 2 - diameter) < threshold:
            return [2 * section.children[0].points[0, COLS.R],
                    2 * section.children[1].points[0, COLS.R]]

    L.warning("Oh no, no similar diameter found!")
    return [original_section.points[-1, COLS.R] * 2,
            original_section.points[-1, COLS.R] * 2]


def _branching_angles(section, order_offset=0):
    '''Return a list of 2-tuples. The first element is the branching order and the second one is
    the angles between the direction of the section and its children's ones

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/morphstats.cpp#n194  # noqa, pylint: disable=line-too-long
    '''
    if section_length(section) < EPSILON:
        return []
    res = []

    branching_order = branch_order(section) - order_offset
    for child in section.children:
        if section_length(child) < EPSILON:
            continue

        theta = np.math.acos(np.dot(direction(section), direction(child)) /
                             (section_length(section) * section_length(child)))
        res.append((branching_order, theta))
    return res


def _continuation(sec, origin):
    '''Continue growing the section

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n241  # noqa, pylint: disable=line-too-long
    '''
    # The following lines is from BlueRepairSDK's code but I'm not
    # convinced by its relevance
    # if first_sec:
    #     direction_ = sec.points[-1, COLS.XYZ] - sec.points[0, COLS.XYZ]
    # else:
    section_direction = _last_segment_vector(sec)
    section_direction /= np.linalg.norm(section_direction)

    radial_direction = sec.points[0, COLS.XYZ] - origin
    length = np.linalg.norm(radial_direction)
    if length > EPSILON:
        radial_direction /= length
    else:
        radial_direction = sec.points[1, COLS.XYZ] - sec.points[0, COLS.XYZ]
        radial_direction /= np.linalg.norm(radial_direction)

    noise_direction = (2 * np.random.random(size=3) - 1)

    direction_ = section_direction + \
        SOMA_REPULSION * radial_direction + \
        NOISE_CONTINUATION * noise_direction

    direction_ /= np.linalg.norm(direction_)

    coord = sec.points[-1, COLS.XYZ] + direction_ * SEG_LENGTH
    radius = np.mean(sec.points[:, COLS.R])
    new_point = np.append(coord, radius)
    sec.points = np.vstack((sec.points, new_point))


def _y_cylindrical_extent(section):
    '''Returns the distance from the last section point to the origin in the XZ plane'''
    xz_last_point = section.points[-1, [0, 2]]
    return np.linalg.norm(xz_last_point)


def _max_y_dendritic_cylindrical_extent(neuron):
    '''Return the maximum distance of dendritic section ends and the origin in the XZ plane'''
    return max(_y_cylindrical_extent(section) for section in neuron.iter()
               if section.type in {SectionType.basal_dendrite, SectionType.apical_dendrite})


class Repair(object):
    '''The repair class'''

    def __init__(self, inputfile, axons=None, seed=0, plane=None):
        '''Repair the input morphology

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/repair.cpp#n469  # noqa, pylint: disable=line-too-long
        '''
        np.random.seed(seed)
        if plane is None:
            L.info('No cut plane specified. Calling CutPlane.find...')
            cut_plane = CutPlane.find(inputfile, bin_width=15)
            L.info('Found cut plane: %s', cut_plane)
        else:
            if isinstance(plane, CutPlane):
                cut_plane = plane
            else:
                cut_plane = CutPlane.from_json(plane)

        self.inputfile = inputfile
        self.axon_donors = axons or list()
        self.donated_intact_axon_sections = list()
        self.cut_leaves = cut_plane.cut_leaves_coordinates
        self.neuron = load_neuron(inputfile)
        self.repair_type_map = dict()
        self.max_y_cylindrical_extent = _max_y_dendritic_cylindrical_extent(self.neuron)
        self.max_y_extent = max(np.max(section.points[:, COLS.Y])
                                for section in self.neuron.iter())

        self.info = dict()
        apical_section_id, _ = apical_point_section_segment(self.neuron)
        if apical_section_id:
            self.apical_section = self.neuron.sections[apical_section_id]
        else:
            self.apical_section = None

    def run(self, outputfile, plot_file=None):
        '''Run'''
        if self.cut_leaves.size == 0:
            L.warning('No cut leaves. Nothing to repair for morphology %s', self.inputfile)
            self.neuron.write(outputfile)
            return

        # The only purpose of the 'planes' variable is to keep
        # each plane.morphology alive. Otherwise sections become unusable
        # https://github.com/BlueBrain/MorphIO/issues/29
        planes = []
        for axon_donor in self.axon_donors:
            plane = CutPlane.find(axon_donor)
            planes.append(plane)
            no_cut_plane = (plane.minus_log_prob < 50)
            self.donated_intact_axon_sections.extend(
                [section for section in iter_sections(plane.morphology)
                 if section.type == SectionType.axon and
                 (no_cut_plane or is_branch_intact(section, plane.cut_leaves_coordinates))])

        self._fill_repair_type_map()
        self._fill_statistics_for_intact_subtrees()
        intact_axonal_sections = [section for section in iter_sections(self.neuron)
                                  if section.type == SectionType.axon and
                                  is_branch_intact(section, self.cut_leaves)]

        # BlueRepairSDK used to have a bounding cylinder filter but
        # I don't know what is it good at so I have commented
        # the only relevant line
        # bounding_cylinder_radius = 10000
        cut_sections_in_bounding_cylinder = [
            section for section in iter_sections(self.neuron)
            if (is_cut_section(section, cut_points=self.cut_leaves)
                # and np.linalg.norm(section.points[-1, COLS.XZ]) < bounding_cylinder_radius
                )
        ]

        cut_leaves_ids = {section: len(section.points)
                          for section in cut_sections_in_bounding_cylinder}

        for section in sorted(cut_sections_in_bounding_cylinder, key=section_path_length):
            type_ = self.repair_type_map[section]
            L.info('Repairing: %s, %s', type_, section.id)
            if type_ in {RepairType.basal, RepairType.oblique, RepairType.tuft}:
                origin = self._get_origin(section)
                if section.type == NeuriteType.basal_dendrite:
                    _continuation(section, origin)
                self._grow(section, self._get_order_offset(section), origin)
            elif type_ == RepairType.axon:
                axon.repair(self.neuron, section, intact_axonal_sections,
                            self.donated_intact_axon_sections, self.max_y_extent)
            elif type_ == RepairType.trunk:
                L.info('Trunk repair is not (nor has ever been) implemented')
            else:
                raise Exception('Unknown type: {}'.format(type_))

        if plot_file is not None:
            try:
                from neuror.view import plot_repaired_neuron
                plot_repaired_neuron(self.neuron, cut_leaves_ids, plot_file)
            except ImportError:
                L.warning('Skipping writing plots as [plotly] extra is not installed')

        self.neuron.write(outputfile)
        L.info('Repair successful for %s', self.inputfile)

    def _find_intact_obliques(self):
        '''
        Find root sections of all intact obliques

        Root obliques are obliques with a section parent of type 'trunk'

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n193  # noqa, pylint: disable=line-too-long
        '''
        root_obliques = (section for section in iter_sections(self.neuron)
                         if (self.repair_type_map[section] == RepairType.oblique and
                             not section.is_root() and
                             self.repair_type_map[section.parent] == RepairType.trunk))
        intacts = [oblique for oblique in root_obliques
                   if is_branch_intact(oblique, self.cut_leaves)]
        return intacts

    def _find_intact_sub_trees(self):
        '''Returns intact neurites'''
        basals = [neurite.root_node for neurite in iter_neurites(self.neuron)
                  if (neurite.type == NeuriteType.basal_dendrite and
                      is_branch_intact(neurite.root_node, self.cut_leaves))]

        if not basals:
            raise Exception('No intact basal dendrites !')

        axons = [neurite.root_node for neurite in iter_neurites(self.neuron)
                 if (neurite.type == NeuriteType.axon and
                     is_branch_intact(neurite.root_node, self.cut_leaves))]
        obliques = self._find_intact_obliques()

        tufts = [section for section in iter_sections(self.neuron)
                 if (self.repair_type_map[section] == RepairType.tuft and
                     not is_cut_section(section, self.cut_leaves))]

        return basals + obliques + axons + tufts

    def _intact_branching_angles(self, branches):
        '''
        Returns lists of branching angles stored in a nested dict
        1st key: section type, 2nd key: branching order

        Args:
            branches (List[Neurite])

        Returns:
            Dict[SectionType, Dict[int, List[int]]]: Branching angles
        '''
        res = defaultdict(lambda: defaultdict(list))
        for branch in branches:
            order_offset = self._get_order_offset(branch)
            for section in branch.ipreorder():
                for order, angle in _branching_angles(section, order_offset):
                    res[self.repair_type_map[section]][order].append(angle)
        return dict(res)

    def _best_case_angle_data(self, section_type, branching_order):
        '''Get the distribution of branching angles for this section type and
        branching order

        If no data are available, fallback on aggregate data

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n329  # noqa, pylint: disable=line-too-long
        '''
        angles = self.info['intact_branching_angles'][section_type]
        accurate_data = angles[branching_order]
        if accurate_data:
            return accurate_data

        return list(chain.from_iterable(angles.values()))

    def _get_origin(self, branch):
        '''Return what should be considered as the origin for this branch'''
        if self.repair_type_map[branch] == RepairType.oblique:
            return branch.points[0, COLS.XYZ]
        if self.repair_type_map[branch] == RepairType.tuft:
            return self.apical_section.points[-1, COLS.XYZ]
        return self.neuron.soma.center

    def _get_order_offset(self, branch):
        r'''
        Return what should be considered as the branch order offset for this branch

        For obliques, the branch order is computed with respect to
        the branch order of the first oblique section so we have to remove the offset.

                    3
                 2  |
     oblique ---->\ |
                   \| 1
                    |
                    |
                    | 0

        oblique order is 2 - 1 = 1
        '''
        if self.repair_type_map[branch] == RepairType.oblique:
            return branch_order(branch)
        if self.repair_type_map[branch] == RepairType.tuft:
            return branch_order(self.apical_section)
        return 0

    def _compute_sholl_data(self, branches):
        '''Compute the number of termination, bifurcation and continuation section for each
        neurite type, sholl layer and shell order

        data[neurite_type][layer][order][action_type] = counts

        Args:
            branches: a collection of Neurite or Section that will be traversed

        Note: This is based on
        https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/morphstats.cpp#n93
        '''
        data = defaultdict(lambda: defaultdict(dict))

        for branch in branches:
            origin = self._get_origin(branch)
            order_offset = self._get_order_offset(branch)

            for section in branch.ipreorder():
                repair_type = self.repair_type_map[section]
                assert repair_type == self.repair_type_map[branch], \
                    'RepairType should not change along the branch way'
                order = branch_order(section) - order_offset
                first_layer, last_layer = (np.linalg.norm(
                    section.points[[0, -1], COLS.XYZ] - origin, axis=1) // SHOLL_LAYER_SIZE).astype(
                        int)
                per_type = data[repair_type]
                for layer in range(min(first_layer, last_layer), max(first_layer, last_layer) + 1):
                    if order not in per_type[layer]:
                        per_type[layer][order] = {Action.TERMINATION: 0,
                                                  Action.CONTINUATION: 0,
                                                  Action.BIFURCATION: 0}
                    per_type[layer][order][Action.CONTINUATION] += 1

                per_type[last_layer][order][Action.BIFURCATION if section.children else
                                            Action.TERMINATION] += 1
        return data

    def _bifurcation(self, section, order_offset):
        '''Create 2 children at the end of the current section

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n287  # noqa, pylint: disable=line-too-long
        '''
        current_diameter = section.points[-1, COLS.R] * 2
        child_diameters = _get_similar_child_diameters(self.info['dendritic_sections'], section)

        median_angle = np.median(
            self._best_case_angle_data(self.repair_type_map[section],
                                       branch_order(section) - order_offset))

        last_segment_vec = _last_segment_vector(section)
        orthogonal = np.cross(last_segment_vec, section.points[-1, COLS.XYZ])

        def shuffle_direction(direction_):
            '''Return the direction to which to grow a child section'''
            theta = median_angle + np.radians(np.random.random() * BIFURCATION_ANGLE)
            first_rotation = np.dot(rotation_matrix(orthogonal, theta), direction_)
            new_dir = np.dot(rotation_matrix(direction_, np.random.random() * np.pi * 2),
                             first_rotation)
            return new_dir / np.linalg.norm(new_dir)

        for child_diameter in child_diameters:
            child_start = section.points[-1, COLS.XYZ]
            points = [child_start.tolist(),
                      (child_start + shuffle_direction(last_segment_vec) * SEG_LENGTH).tolist()]
            prop = PointLevel(points, [current_diameter, child_diameter])

            child = section.append_section(prop)
            L.debug('section appended: %s', child.id)

    def _grow(self, section, order_offset, origin):
        '''grow main method

        Will either:
            - continue growing the section
            - create a bifurcation
            - terminate the growth

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n387  # noqa, pylint: disable=line-too-long
        '''
        if (self.repair_type_map[section] == RepairType.tuft and
                _y_cylindrical_extent(section) > self.max_y_cylindrical_extent):
            return

        sholl_layer = _get_sholl_layer(section, origin)
        pseudo_order = branch_order(section) - order_offset
        L.debug('In _grow. Layer: %s, order: %s', sholl_layer, pseudo_order)

        proba = _get_sholl_proba(self.info['sholl'],
                                 self.repair_type_map[section],
                                 sholl_layer,
                                 pseudo_order)

        L.debug('action proba[%s][%s][%s]: %s', section.type, sholl_layer, pseudo_order, proba)
        action = np.random.choice(list(proba.keys()), p=list(proba.values()))

        if action == Action.CONTINUATION:
            L.info('Continuing')
            backwards_sections = _grow_until_sholl_sphere(section, origin, sholl_layer)

            if backwards_sections == 0:
                self._grow(section, order_offset, origin)
            L.debug(section.points[-1])

        elif action == Action.BIFURCATION:
            L.info('Bifurcating')
            backwards_sections = _grow_until_sholl_sphere(section, origin, sholl_layer)
            if backwards_sections == 0:
                self._bifurcation(section, order_offset)
                for child in section.children:
                    self.repair_type_map[child] = self.repair_type_map[section]
                    self._grow(child, order_offset, origin)

    def _fill_statistics_for_intact_subtrees(self):
        '''Compute statistics'''
        branches = self._find_intact_sub_trees()

        self.info = dict(
            intact_branching_angles=self._intact_branching_angles(branches),
            dendritic_sections=[section for section in iter_sections(self.neuron)
                                if section.type in {nm.APICAL_DENDRITE, nm.BASAL_DENDRITE}],
            sholl=self._compute_sholl_data(branches),
        )

    def _fill_repair_type_map(self):
        '''Assign a repair section type to each section

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/repair.cpp#n242  # noqa, pylint: disable=line-too-long
        '''
        for section in iter_sections(self.neuron):
            if section.type == SectionType.apical_dendrite:
                self.repair_type_map[section] = RepairType.oblique
            elif section.type == SectionType.basal_dendrite:
                self.repair_type_map[section] = RepairType.basal
            elif section.type == SectionType.axon:
                self.repair_type_map[section] = RepairType.axon

        if self.apical_section is not None:
            for section in self.apical_section.ipreorder():
                self.repair_type_map[section] = RepairType.tuft

            # The value for the apical section must be overriden to 'trunk'
            for section in self.apical_section.iupstream():
                self.repair_type_map[section] = RepairType.trunk


def repair(inputfile, outputfile, axons=None, seed=0, plane=None, plot_file=None):
    '''The repair function'''
    if axons is None:
        axons = list()
    obj = Repair(inputfile, axons=axons, seed=seed, plane=plane)
    obj.run(outputfile, plot_file=plot_file)



from pathlib import Path
from neuror.sanitize import fix_non_zero_segments, sanitize, sanitize_all, CorruptedMorphology

from morphio import Morphology
from numpy.testing import assert_equal, assert_array_equal

from nose.tools import assert_raises
from tempfile import TemporaryDirectory

PATH = Path(__file__).parent / 'data'

def test_fix_non_zero_segments():
    neuron = fix_non_zero_segments(Path(PATH, 'simple-with-duplicates.asc'))
    assert_equal(len(neuron.root_sections), 1)
    assert_array_equal(neuron.section(0).points,
                       [[0., 0., 0.],
                        [1., 1., 0.],
                        [2., 0., 0.],
                        [3., 0., 0.]])


def test_sanitize():
    with TemporaryDirectory('test-sanitize') as tmp_folder:
        output_path = Path(tmp_folder, 'sanitized.asc')
        sanitize(Path(PATH, 'simple-with-duplicates.asc'), output_path)
        neuron = Morphology(output_path)
        assert_equal(len(neuron.root_sections), 1)
        assert_array_equal(neuron.section(0).points,
                           [[0., 0., 0.],
                            [1., 1., 0.],
                            [2., 0., 0.],
                            [3., 0., 0.]])

        assert_raises(CorruptedMorphology, sanitize,
                      Path(PATH, 'no-soma.asc'), Path(tmp_folder, 'no-soma.asc'))

        assert_raises(CorruptedMorphology, sanitize,
                      Path(PATH, 'negative-diameters.asc'), Path(tmp_folder, 'negative-diameter.asc'))


def test_sanitize_all():
    with TemporaryDirectory('test-sanitize') as tmp_folder:
        tmp_folder = Path(tmp_folder)
        sanitize_all(PATH / 'input-sanitize-all', tmp_folder)

        assert_array_equal(list(sorted(tmp_folder.rglob('*.asc'))),
                           [tmp_folder / 'a.asc',
                            tmp_folder / 'sub-folder/sub-sub-folder/c.asc'])



'''Module for the sanitization of raw morphologies.'''
import logging
from pathlib import Path
from tqdm import tqdm

import numpy as np

import morphio
from morphio import MorphioError, set_maximum_warnings, SomaType
from morphio.mut import Morphology  # pylint: disable=import-error

L = logging.getLogger('neuror')


class CorruptedMorphology(Exception):
    '''Exception for morphologies that should not be used'''


def iter_morphologies(folder):
    '''Recursively yield morphology files in folder and its sub-directories.'''
    return (path for path in folder.rglob('*') if path.suffix.lower() in {'.swc', '.h5', '.asc'})


def sanitize(input_neuron, output_path):
    '''Sanitize one morphology.

    - fixes non zero segments
    - raises if the morphology has no soma
    - raises if the morphology has negative diameters

    Args:
        input_neuron (str|pathlib.Path|morphio.Morphology|morphio.mut.Morphology): input neuron
        output_path (str|pathlib.Path): output name
    '''
    neuron = morphio.Morphology(input_neuron)
    if neuron.soma.type == SomaType.SOMA_UNDEFINED:  # pylint: disable=no-member
        raise CorruptedMorphology('{} has no soma'.format(input_neuron))
    if np.any(neuron.diameters < 0):
        raise CorruptedMorphology('{} negative diameters'.format(input_neuron))

    fix_non_zero_segments(neuron).write(str(output_path))


def sanitize_all(input_folder, output_folder):
    '''Sanitize all morphologies in input_folder and its sub-directories.

    Note: the sub-directory structure is maintained.

    - fixes non zero segments
    - raises if the morphology has no soma
    - raises if the morphology has negative diameters

    Args:
        input_folder (str|pathlib.Path): input neuron
        output_folder (str|pathlib.Path): output name
    '''
    set_maximum_warnings(0)

    errored_paths = list()
    for path in tqdm(list(iter_morphologies(Path(input_folder)))):
        relative_path = path.relative_to(input_folder)
        output_dir = output_folder / relative_path.parent
        if not output_dir.exists():
            output_dir.mkdir(parents=True)
        try:
            sanitize(path, output_dir / path.name)
        except (MorphioError, CorruptedMorphology):
            errored_paths.append(str(path))
    if errored_paths:
        L.info('Files in error:')
        for path in errored_paths:
            L.info(path)


def fix_non_zero_segments(neuron):
    '''Return a neuron with zero length segments removed

    Sections composed of a single zero length segment are deleted
    Args:
        neuron (str|pathlib.Path|morphio.Morphology|morphio.mut.Morphology): input neuron

    Returns:
        a fixed morphio.mut.Morphology
    '''
    neuron = Morphology(neuron)
    to_be_deleted = list()
    for section in neuron.iter():
        points = section.points
        distances = np.linalg.norm(np.diff(points, axis=0), axis=1)
        indices = np.append(0, np.nonzero(distances)[0] + 1)
        if len(indices) != len(points):
            section.points = section.points[indices]
            section.diameters = section.diameters[indices]
        if len(indices) < 2:
            to_be_deleted.append(section)

    for section in to_be_deleted:
        neuron.delete_section(section)
    return neuron



'''
The module to run the full repair
'''
import logging
import os
from pathlib import Path

from morph_tool.utils import iter_morphology_files

from neuror.main import repair
from neuror.unravel import unravel_all

L = logging.getLogger('neuror')


def _get_folders(root_dir,
                 raw_dir=None,
                 raw_planes_dir=None,
                 unravelled_dir=None,
                 unravelled_planes_dir=None,
                 repaired_dir=None,
                 plots_dir=None):
    '''Get folder paths with sensible defaults'''
    folders = dict()
    folders['raw'] = raw_dir or str(Path(root_dir, 'raw'))
    folders['raw_planes'] = raw_planes_dir or str(Path(folders['raw'], 'planes'))
    folders['unravelled'] = unravelled_dir or str(Path(root_dir, 'unravelled'))
    folders['unravelled_planes'] = unravelled_planes_dir or str(
        Path(folders['unravelled'], 'planes'))
    folders['repaired'] = repaired_dir or str(Path(root_dir, 'repaired'))
    folders['plots'] = plots_dir or str(Path(root_dir, 'plots'))
    return folders


def repair_all(input_dir, output_dir, seed=0, axons=None, planes_dir=None, plots_dir=None):
    '''Repair all morphologies in input folder'''
    for f in iter_morphology_files(input_dir):
        L.info(f)
        inputfilename = Path(input_dir, f)
        outfilename = Path(output_dir, os.path.basename(f))
        if planes_dir:
            plane = str(Path(planes_dir, inputfilename.name).with_suffix('.json'))
        else:
            plane = None

        if plots_dir is not None:
            name = 'neuron_{}.html'.format(Path(inputfilename).stem.replace(' ', '_'))
            plot_file = str(Path(plots_dir, name))
        else:
            plot_file = None

        try:
            repair(str(inputfilename), str(outfilename),
                   seed=seed, axons=axons, plane=plane, plot_file=plot_file)
        except Exception as e:  # noqa, pylint: disable=broad-except
            L.warning('%s failed', f)
            L.warning(e, exc_info=True)


# pylint: disable=too-many-arguments
def full(root_dir,
         seed=0,
         window_half_length=5,
         raw_dir=None,
         raw_planes_dir=None,
         unravelled_dir=None,
         unravelled_planes_dir=None,
         repaired_dir=None,
         plots_dir=None):
    '''
    Perform the unravelling and repair in ROOT_DIR:

    1) perform the unravelling of the neuron
    2) update the cut points position after unravelling and writes it
       in the unravelled/planes folder
    3) repair the morphology

    All output directories can be overriden with the corresponding arguments.
    Here is the default structure:

    - raw_dir: ROOT_DIR/raw/ with all raw morphologies to repair
    - raw_planes_dir: RAW_DIR/planes with all cut planes
    - unravelled_dir: ROOT_DIR/unravelled/ where unravelled morphologies will be written
    - unravelled_planes_dir: UNRAVELLED_DIR/planes where unravelled planes will be written
    - repaired_dir: ROOT_DIR/repaired/ where repaired morphologies will be written
    - plots_dir: ROOT_DIR/plots where the plots will be put
    '''

    folders = _get_folders(root_dir,
                           raw_dir,
                           raw_planes_dir,
                           unravelled_dir,
                           unravelled_planes_dir,
                           repaired_dir,
                           plots_dir)

    for folder in ['raw', 'raw_planes']:
        if not os.path.exists(folders[folder]):
            raise Exception('%s does not exists' % folders[folder])

    for folder in ['unravelled', 'unravelled_planes', 'repaired', 'plots']:
        if not os.path.exists(folders[folder]):
            os.mkdir(folders[folder])

    unravel_all(folders['raw'],
                folders['unravelled'],
                folders['raw_planes'],
                folders['unravelled_planes'],
                window_half_length=window_half_length)
    repair_all(folders['unravelled'],
               folders['repaired'],
               seed=seed,
               planes_dir=folders['unravelled_planes'],
               plots_dir=folders['plots'])

    try:
        from neuror.view import view_all
        view_all([folders['raw'], folders['unravelled'], folders['repaired']],
                 titles=['raw', 'unravelled', 'repaired'],
                 output_pdf=str(Path(folders['plots'], 'report.pdf')))
    except ImportError:
        L.warning('Skipping writing plots as [plotly] extra is not installed')



import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from nose.tools import assert_equal
from click.testing import CliRunner

from neuror.cli import cli

DATA = Path(__file__).parent / 'data'


def test_cli():
    runner = CliRunner()

    with TemporaryDirectory('test-file') as folder:
        result = runner.invoke(cli, ['cut-plane', 'repair', 'file',
                                     str(DATA / 'real.asc'),
                                     '/tmp/test_repair_cli.asc'])
        assert_equal(result.exit_code, 0)

    with TemporaryDirectory('test-cli-folder') as folder:
        result = runner.invoke(cli, ['cut-plane', 'repair', 'folder', str(DATA), folder])
        assert_equal(result.exit_code, 0)


def test_cli_full():
    runner = CliRunner()
    with TemporaryDirectory('test-cli-full') as tmp_folder:
        test_folder = str(Path(tmp_folder, 'test-full-repair'))
        shutil.copytree(DATA / 'test-full-repair', test_folder)
        result = runner.invoke(cli, ['cut-plane', 'repair', 'full', test_folder])
        assert_equal(result.exit_code, 0)


def test_cli_axon():
    runner = CliRunner()
    with TemporaryDirectory('test-cli-axon') as tmp_folder:
        tmp_folder = Path(tmp_folder)
        result = runner.invoke(cli, ['cut-plane', 'repair', 'file',
                                     '-a', str(DATA / 'real-with-axon.asc'),
                                     str(DATA / 'real-with-axon.asc'),
                                     str(tmp_folder / 'output.asc')])
        assert_equal(result.exit_code, 0)


def test_sanitize():
    runner = CliRunner()
    with TemporaryDirectory('test-cli-axon') as tmp_folder:
        tmp_folder = Path(tmp_folder)
        result = runner.invoke(cli, ['sanitize', 'file',
                                     str(DATA / 'simple-with-duplicates.asc'),
                                     str(tmp_folder / 'output.asc')])
        assert_equal(result.exit_code, 0)

        result = runner.invoke(cli, ['sanitize', 'folder',
                                    str(DATA), str(tmp_folder)])
        assert_equal(result.exit_code, 0)



'''Fix zero diameters

Re-implementation of: https://bbpcode.epfl.ch/source/xref/sim/MUK/apps/Fix_Zero_Diameter.cpp
'''
import sys
from collections import namedtuple

import numpy as np

SMALL = 0.0001

Point = namedtuple('Point',
                   ['section',
                    'point_id'])  # index of the point within the given section


def fix_neurite(root_section):
    '''Apply all fixes to a neurite'''
    point = Point(root_section, 0)
    fix_from_downstream(point)
    fix_in_between(point, stack=list())
    fix_from_upstream(point, 0)


def next_section_points(point):
    '''Yields next point along the section or the starting points
    of the child sections

    This function mimics the traversal of the linked list returned by recurseSection() at
    https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#40

    It is what enables this reimplementation to use the same algorithm as the original code
    '''
    size = len(point.section.points)
    if point.point_id < size - 1:
        yield Point(point.section, point.point_id + 1)
    else:
        for child in point.section.children:
            yield Point(child, 0)


def next_section_points_upstream(point):
    '''Yield upstream points until reaching the root'''
    section, point_id = point
    while not (section.is_root and point_id == 0):
        if point_id > 0:
            point_id -= 1
        else:
            section = section.parent
            assert len(section.points)
            point_id = len(section.points) - 1
        yield Point(section, point_id)


def set_diameter(point, new_diameter):
    '''Set a given diameter

    Unfortunately morphio does not support single value assignment
    so one has to update the diameters for the whole sections
    '''
    diameters = point.section.diameters
    diameters[point.point_id] = new_diameter
    point.section.diameters = diameters


def fix_from_downstream(point):
    '''Re-implementation of recursePullFix() available at
    https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#66

    If the current diameter is below the threshold, change its value to the biggest value
    among the 1-degree children downstream diameters

    This is recursive only until a diameter above threshold is found so this
    fixes zero diameters that are located at the beginning of the neurite

    Fixes this:
    Soma--0--0--0--1--1--1
    But not this:
    Soma--1--1--1--0--0--0--1--1
    '''
    diameter = point.section.diameters[point.point_id]

    is_leaf = not list(next_section_points(point))
    if diameter > SMALL or is_leaf:
        return diameter

    biggest_diameter_downstream = max(fix_from_downstream(next_point)
                                      for next_point in next_section_points(point))

    set_diameter(point, biggest_diameter_downstream)
    return biggest_diameter_downstream


def fix_from_upstream(point, upstream_good_diameter):
    '''Re-implementation of recursePushFix() available at
    https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#94

    Reset the diameter to upstream_good_diameter if the current value and all child values
    are below threshold

    Args:
        point: the current point
        upstream_good_diameter: the diameter value coming from upstream that will be used if
            the current diameter is not suitable

    Returns:
        The current diameter if above threshold, else the smallest child value above threshold
        else (if no child value is above threshold) returns 0
    '''
    section, point_idx = point
    diameter = section.diameters[point_idx]
    if diameter > SMALL:
        for next_point in next_section_points(point):
            fix_from_upstream(next_point, diameter)
        return diameter

    next_diameters = [fix_from_upstream(next_point, upstream_good_diameter)
                      for next_point in next_section_points(point)]

    smallest = min(next_diameters) if next_diameters else 0

    if smallest == 0:
        set_diameter(point, upstream_good_diameter)
    return smallest


def fix_in_between(point, stack):
    '''Re-implementation of
    https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#162

    Fix diameters between two points with valid diameters by applying a ramp

    Args:
        point: the current points
        stack: the stack of upstream points
    '''
    stack.append(point)
    # pylint: disable=chained-comparison
    # Condition: current point is good and previous point has a zero diameter
    if(len(stack) > 1 and
       stack[-2].section.diameters[stack[-2].point_id] <= SMALL and
       stack[-1].section.diameters[stack[-1].point_id] > SMALL):

        for next_point in next_section_points_upstream(point):
            if next_point.section.diameters[next_point.point_id] > SMALL:
                connect_average(point, next_point)
                break

    for next_point in next_section_points(point):
        fix_in_between(next_point, stack)
    stack.pop()


def connect_average(point1, point2):
    '''Apply a ramp diameter between the two points

    Re-implementation of https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#232
    Contrary to the previous implementation the diameter the ramp is computed in term of
    pathlength and no longer in term of point number
    '''
    sec1, idx1 = point1
    sec2, idx2 = point2
    diam1 = sec1.diameters[idx1]
    diam2 = sec2.diameters[idx2]
    prev_point = point1
    pathlengths = list()

    for point in next_section_points_upstream(point1):
        pathlengths.append(np.linalg.norm(prev_point.section.points[prev_point.point_id] -
                                          point.section.points[point.point_id]))
        prev_point = point
        if point == point2:
            break

    cumulative_pathlengths = np.cumsum(pathlengths)
    pathlength_fractions = cumulative_pathlengths / cumulative_pathlengths[-1]

    for point, fraction in zip(next_section_points_upstream(point1), pathlength_fractions[:-1]):
        set_diameter(point, diam1 + (diam2 - diam1) * fraction)


def fix_zero_diameters(neuron):
    '''Fix zero diameters'''
    old_limit = sys.getrecursionlimit()
    point_counts = sum(len(section.points) for section in neuron.iter())
    sys.setrecursionlimit(max(old_limit, point_counts))
    try:
        for root in neuron.root_sections:
            fix_neurite(root)
    finally:
        sys.setrecursionlimit(old_limit)



"""neuror setup.py"""
import sys

if sys.version_info < (3, 6):
    sys.exit("Sorry, Python < 3.6 is not supported")

import imp
from pathlib import Path

from setuptools import setup, find_packages

VERSION = imp.load_source("neuror.version", "neuror/version.py").VERSION

# read the contents of the README file
readme_path = Path(__file__).resolve().parent / "README.rst"
with open(readme_path, encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='NeuroR',
    author='bbp-ou-nse',
    author_email="bbp-ou-nse@groupes.epfl.ch",
    version=VERSION,
    description='A morphology repair tool',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/bluebrain/neuror",
    license="LGPLv3",
    install_requires=[
        'click>=0.7.0',
        'matplotlib>=2.2.3',
        'morph-tool>=0.1.14',
        'morphio>=2.1.1',
        'neurom @ git+https://git@github.com/BlueBrain/NeuroM.git@mut_morphio#egg=neurom-2.0.0',
        'pandas>=0.24.2',
        'pyquaternion>=0.9.2',
        'scipy>=1.2.0',
    ],
    extras_require={
        'plotly': [
            'dash-core-components>=0.46.0',  # HTML components
            'dash-table>=3.6.0',  # Interactive DataTable component (new!)
            'dash>=0.41.0',  # The core dash backend
            'plotly-helper>=0.0.2',
        ],
        'docs': ['sphinx', 'sphinx-click', 'sphinx_rtd_theme'],
    },
    packages=find_packages(),
    entry_points={'console_scripts': ['neuror=neuror.cli:cli']},
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)



'''The unravel module

Unravelling is the action of "stretching" the cell that
has been shrunk because of the dehydratation caused by the slicing'''
import json
import logging
import os
from pathlib import Path

import morphio
import numpy as np
import pandas as pd
from morph_tool.utils import iter_morphology_files
from scipy.spatial.ckdtree import cKDTree

from neuror.cut_plane import CutPlane
from neuror.utils import RepairJSON

L = logging.getLogger('neuror')

DEFAULT_WINDOW_HALF_LENGTH = 5


def _get_principal_direction(points):
    '''Return the principal direction of a point cloud
    It is the eigen vector of the covariance matrix with the highest eigen value'''

    X = np.copy(np.asarray(points))
    X -= np.mean(X, axis=0)
    C = np.dot(X.T, X)
    w, v = np.linalg.eig(C)
    return v[:, w.argmax()]


def _unravel_section(sec, new_section, window_half_length):
    '''Unravel a section'''
    points = sec.points
    point_count = len(points)
    if new_section.is_root:
        unravelled_points = [new_section.points[0]]
    else:
        unravelled_points = [new_section.parent.points[-1]]

    for window_center in range(1, point_count):
        window_start = max(0, window_center - window_half_length - 1)
        window_end = min(point_count, window_center + window_half_length + 1)

        direction = _get_principal_direction(points[window_start:window_end])

        segment = points[window_center] - points[window_center - 1]

        # make it span length the same as the original segment within the window
        direction *= np.linalg.norm(segment) / np.linalg.norm(direction)

        # point it in the same direction as before
        direction *= np.sign(np.dot(segment, direction))

        unravelled_points.append(direction + unravelled_points[window_center - 1])

    new_section.points = unravelled_points


def unravel(filename, window_half_length=DEFAULT_WINDOW_HALF_LENGTH):
    '''Return an unravelled neuron

    Segment are unravelled iteratively
    Each segment direction is replaced by the averaged direction in a sliding window
    around this segment. And the original segment length is preserved.
    The start position of the new segment is the end of the latest unravelled segment

    Args:
        filename (str): the neuron to unravel
        window_half_length (int): the number of segments that defines half of the sliding window

    Returns:
        a tuple (morphio.mut.Morphology, dict) where first item is the unravelled
            morphology and the second one is the mapping of each point coordinate
            before and after unravelling
    '''
    morph = morphio.Morphology(filename)
    new_morph = morphio.mut.Morphology(morph)  # pylint: disable=no-member

    coord_before = np.empty([0, 3])
    coord_after = np.empty([0, 3])

    for sec, new_section in zip(morph.iter(), new_morph.iter()):
        _unravel_section(sec, new_section, window_half_length)

        coord_before = np.append(coord_before, sec.points, axis=0)
        coord_after = np.append(coord_after, new_section.points, axis=0)

    mapping = pd.DataFrame({
        'x0': coord_before[:, 0],
        'y0': coord_before[:, 1],
        'z0': coord_before[:, 2],
        'x1': coord_after[:, 0],
        'y1': coord_after[:, 1],
        'z1': coord_after[:, 2],
    })

    L.info('Unravel successful for file: %s', filename)
    return new_morph, mapping


def unravel_plane(input_plane, mapping):
    '''Return a new CutPlane object where the cut-leaves
    position has been updated after unravelling'''
    plane = CutPlane.from_json(input_plane)
    leaves = plane.cut_leaves_coordinates
    t = cKDTree(mapping[['x0', 'y0', 'z0']])
    distances, indices = t.query(leaves)
    not_matching_leaves = np.where(distances > 1e-5)[0]
    if not_matching_leaves.size:
        raise Exception('Cannot find the following leaves in the mapping:\n{}'.format(
                        leaves[not_matching_leaves]))
    plane.cut_leaves_coordinates = mapping.iloc[indices][['x1', 'y1', 'z1']].values
    return plane


def unravel_all(raw_dir, unravelled_dir,
                raw_planes_dir,
                unravelled_planes_dir,
                window_half_length=DEFAULT_WINDOW_HALF_LENGTH):
    '''Repair all morphologies in input folder
    '''
    if not os.path.exists(raw_planes_dir):
        raise Exception('{} does not exist'.format(raw_planes_dir))

    if not os.path.exists(unravelled_planes_dir):
        os.mkdir(unravelled_planes_dir)

    for inputfilename in iter_morphology_files(raw_dir):
        L.info('Unravelling: %s', inputfilename)
        outfilename = Path(unravelled_dir, inputfilename.name)
        raw_plane = str(Path(raw_planes_dir, inputfilename.name).with_suffix('.json'))
        unravelled_plane = Path(unravelled_planes_dir, inputfilename.name).with_suffix('.json')

        try:
            neuron, mapping = unravel(str(inputfilename), window_half_length)
            neuron.write(str(outfilename))
            with open(str(unravelled_plane), 'w') as f:
                json.dump(unravel_plane(raw_plane, mapping).to_json(), f, cls=RepairJSON)

        except Exception as e:  # noqa, pylint: disable=broad-except
            L.warning('Unravelling %s failed', f)
            L.warning(e, exc_info=True)
